# Solr redaction instructions and troubleshooting

## Instructions:
Run the following as root to install solr 6.3:
```
cd /opt
wget http://apache.org/dist/lucene/solr/6.3.0/solr-6.3.0.tgz
tar xvfz solr-6.3.0.tgz
useradd -a -G hadoop -G hdfs solr
chown -R solr:solr /opt/solr-6.3.0
ln -s /opt/solr-6.3.0 /opt/solr
```
[Note]	Note
  The useradd -a -G hadoop solr command is required if you are using the HTTP (Spnego) keytab that is generated by Ambari.

  
Create a sym link of the redaction library (su - solr)
```
cd /opt/solr-6.3.0/server/solr-webapp/webapp/WEB-INF/lib
ln -s /opt/pontus/pontus-redaction-hbase-coproc-0.0.1-SNAPSHOT.jar
```
[Note]	Note
  The redaction layer has to modify internal apis within solr, so we have to be loaded with the main classes; otherwise, we start getting exceptions like this:
      44521 WARN  (qtp624271064-29) [c:labs s:shard1 r:core_node1 x:core1] o.e.j.s.ServletHandler Error for /solr/core1/select
    java.lang.IllegalAccessError: tried to access field org.apache.solr.handler.component.ResponseBuilder.requestInfo from class org.apache.solr.handler.component.SolrPontusRequestHandler
            at org.apache.solr.handler.component.SolrPontusRequestHandler.handleRequestBody(SolrPontusRequestHandler.java:204)
            at org.apache.solr.handler.RequestHandlerBase.handleRequest(RequestHandlerBase.java:155)
            at org.apache.solr.core.SolrCore.execute(SolrCore.java:2102)
            at org.apache.solr.servlet.HttpSolrCall.execute(HttpSolrCall.java:654)
            at org.apache.solr.servlet.HttpSolrCall.call(HttpSolrCall.java:460)
            at org.apache.solr.servlet.SolrDispatchFilter.doFilter(SolrDispatchFilter.java:257)
            at org.apache.solr.servlet.SolrDispatchFilter.doFilter(SolrDispatchFilter.java:208)
            at org.eclipse.jetty.servlet.ServletHandler$CachedChain.doFilter(ServletHandler.java:1652)
            at org.eclipse.jetty.servlet.ServletHandler.doHandle(ServletHandler.java:585)
            at org.eclipse.jetty.server.handler.ScopedHandler.handle(ScopedHandler.java:143)
            at org.eclipse.jetty.security.SecurityHandler.handle(SecurityHandler.java:577)
            at org.eclipse.jetty.server.session.SessionHandler.doHandle(SessionHandler.java:223)
            at org.eclipse.jetty.server.handler.ContextHandler.doHandle(ContextHandler.java:1127)
            at org.eclipse.jetty.servlet.ServletHandler.doScope(ServletHandler.java:515)


Create the following file:  /home/solr/solr_jaas.conf (and then a sym link to /opt/solr/solr_jaas.conf
```
Client {
       com.sun.security.auth.module.Krb5LoginModule required
       refreshKrb5Config=true
       useTicketCache=true
       renewTGT=true
       doNotPrompt=true
       useKeyTab=true
       keyTab="/etc/security/keytabs/spnego.service.keytab"
       storeKey=true
       debug=true
       isInitiator=true
       principal="HTTP/sandbox.hortonworks.com@YOUR_REALM_GOES_HERE";

};
KafkaClient {
   com.sun.security.auth.module.Krb5LoginModule required
   refreshKrb5Config=true
   useTicketCache=true
   renewTGT=true
   doNotPrompt=true
   useKeyTab=true
   keyTab="/etc/security/keytabs/spnego.service.keytab"
   storeKey=true
   debug=true
   isInitiator=true
   principal="HTTP/sandbox.hortonworks.com@YOUR_REALM_GOES_HERE"
   serviceName="kafka";
};
```

[Note]  Note
The /etc/security/keytabs/spnego.service.keytab should be created by HDP when you hit the configure Kerberos button in ambari.


Run kadmin.local to add the solr credentials:
```
Authenticating as principal HTTP/admin@YOUR_REALM_GOES_HERE with password.
kadmin.local:  addprinc solr
WARNING: no policy specified for solr@YOUR_REALM_GOES_HERE; defaulting to no policy
Enter password for principal "solr@YOUR_REALM_GOES_HERE":
Re-enter password for principal "solr@YOUR_REALM_GOES_HERE":
Principal "solr@YOUR_REALM_GOES_HERE" created.
kadmin.local:  xst -k /etc/security/keytabs/solr.service.keytab
Usage: ktadd [-k[eytab] keytab] [-q] [-e keysaltlist] [-norandkey] [principal | -glob princ-exp] [...]
kadmin.local:  xst -k /etc/security/keytabs/solr.service.keytab solr
Entry for principal solr with kvno 2, encryption type aes256-cts-hmac-sha1-96 added to keytab WRFILE:/etc/security/keytabs/solr.service.keytab.
Entry for principal solr with kvno 2, encryption type aes128-cts-hmac-sha1-96 added to keytab WRFILE:/etc/security/keytabs/solr.service.keytab.
Entry for principal solr with kvno 2, encryption type des3-cbc-sha1 added to keytab WRFILE:/etc/security/keytabs/solr.service.keytab.
Entry for principal solr with kvno 2, encryption type arcfour-hmac added to keytab WRFILE:/etc/security/keytabs/solr.service.keytab.
```



Add the following lines to the BOTTOM of the /opt/solr/bin/solr.in.sh file:
```
SOLR_JAAS_FILE=/home/solr/solr_jaas.conf
SOLR_HOST=`hostname -f`
ZK_HOST="sandbox.hortonworks.com:2181/solr"
KERBEROS_REALM="YOUR_REALM_GOES_HERE"
SOLR_KEYTAB=/etc/security/keytabs/spnego.service.keytab
SOLR_KERB_PRINCIPAL=HTTP/${SOLR_HOST}@${KERBEROS_REALM}
SOLR_KDC=sandbox.hortonworks.com

#SOLR_AUTHENTICATION_CLIENT_CONFIGURER=
#SOLR_AUTHENTICATION_OPTS=

SOLR_AUTHENTICATION_CLIENT_CONFIGURER="org.apache.solr.client.solrj.impl.Krb5HttpClientConfigurer"
SOLR_AUTHENTICATION_OPTS=" \
 -DauthenticationPlugin=org.apache.solr.security.KerberosPlugin \
 -Djava.security.krb5.realm=${KERBEROS_REALM} \
 -Djava.security.krb5.kdc=${SOLR_KDC} \
 -Djavax.security.auth.useSubjectCredsOnly=false \
 -Djava.security.auth.login.config="$SOLR_JAAS_FILE" \
 -Dsolr.kerberos.principal=${SOLR_KERB_PRINCIPAL} \
 -Dsolr.kerberos.keytab="${SOLR_KEYTAB}"\
 -Dsolr.kerberos.cookie.domain=${SOLR_HOST} \
 -Dhost=${SOLR_HOST} \
 -Dsun.security.spnego.debug=true \
 -Dsun.security.krb5.debug=true \
 -Dsun.security.jgss.debug=true \
 -Dsun.security.spnego.msinterop=true \
 -Dhadoop.security.authentication=kerberos \
 -Dgroup.id=solr_${SOLR_HOST} \
 -Dsolr.kerberos.name.rules=DEFAULT"
```

Set up the Kerberos Plugin in zookeeper:
```
kinit -k -t /etc/security/keytabs/solr.service.keytab solr/sandbox.hortonworks.com
export JVMFLAGS="-Djava.security.auth.login.config=/home/solr/solr-jaas.conf"
/opt/solr/server/scripts/cloud-scripts/zkcli.sh -zkhost localhost:2181 -cmd put /solr/security.json '{"authentication":{"class": "org.apache.solr.security.KerberosPlugin"}}'
```      
      
      
Add/replace the following entries to solrconfig.xml (note that a copy of this should be in the git dir:
```
<config>
  ....
            <directoryFactory name="DirectoryFactory" class="solr.HdfsDirectoryFactory">
                <str name="solr.hdfs.home">hdfs://sandbox.hortonworks.com/user/solr</str>
                <bool name="solr.hdfs.blockcache.enabled">true</bool>
                <int name="solr.hdfs.blockcache.slab.count">1</int>
                <bool name="solr.hdfs.blockcache.direct.memory.allocation">false</bool>
                <int name="solr.hdfs.blockcache.blocksperbank">16384</int>
                <bool name="solr.hdfs.blockcache.read.enabled">true</bool>
                <bool name="solr.hdfs.blockcache.write.enabled">false</bool>
                <bool name="solr.hdfs.nrtcachingdirectory.enable">true</bool>
                <int name="solr.hdfs.nrtcachingdirectory.maxmergesizemb">16</int>
                <int name="solr.hdfs.nrtcachingdirectory.maxcachedmb">192</int>
                <str name="hadoop.security.authentication">Kerberos</str>
                <bool name="solr.hdfs.security.kerberos.enabled">true</bool>
                <str name="solr.hdfs.security.kerberos.keytabfile">/etc/security/keytabs/solr.service.keytab</str>
                <str name="solr.hdfs.security.kerberos.principal">solr@YOUR_REALM_GOES_HERE</str>
                <str name="solr.hdfs.confdir">/usr/hdp/2.5.0.0-1245/hadoop/conf</str>
            </directoryFactory>


  <queryResponseWriter name="xml"
                      default="true"
                      class="uk.gov.homeoffice.pontus.solr.SolrPoleSecurityXMLResponseWriterFilter" />

  <queryResponseWriter name="csv"
                      default="true"
                      class="uk.gov.homeoffice.pontus.solr.SolrPoleSecurityCSVResponseWriterFilter" />

  <queryResponseWriter name="json"
                    default="true"                    class="uk.gov.homeoffice.pontus.solr.SolrPoleSecurityJSONResponseWriterFilter">
                    
      <!-- For the purposes of the tutorial, JSON responses are written as
     plain text so that they are easy to read in *any* browser.
     If you expect a MIME type of "application/json" just remove this override.
    -->
    <str name="content-type">text/plain; charset=UTF-8</str>
  </queryResponseWriter>
  
  <queryParser name="PontusSecurity" class="uk.gov.homeoffice.pontus.solr.SolrPoleSecurityQParserPlugin"/>  
  
    <requestHandler name="/query" class="org.apache.solr.handler.component.SolrPontusRequestHandler">
    <lst name="defaults">
      <str name="echoParams">explicit</str>
      <str name="wt">json</str>
      <str name="indent">true</str>
    </lst>
  </requestHandler>

  <requestHandler name="/browse" class="org.apache.solr.handler.component.SolrPontusRequestHandler" useParams="query,facets,velocity,browse">
    <lst name="defaults">
      <str name="echoParams">explicit</str>
    </lst>
  </requestHandler>
  
  <requestHandler name="/export" class="org.apache.solr.handler.component.SolrPontusRequestHandler">
    <lst name="invariants">
      <str name="rq">{!xport}</str>
      <str name="wt">xsort</str>
      <str name="distrib">false</str>
    </lst>

    <arr name="components">
      <str>query</str>
    </arr>
  </requestHandler>                    
  
</config>
```
  
Install the solrconfig.xml in zk:
```
/opt/solr/server/scripts/cloud-scripts/zkcli.sh -zkhost localhost:2181/solr -cmd makepath core1
/opt/solr/server/scripts/cloud-scripts/zkcli.sh -zkhost localhost:2181/solr -cmd makepath core1/data
/opt/solr/server/scripts/cloud-scripts/zkcli.sh -zkhost localhost:2181/solr -cmd putfile /configs/labs/solrconfig.xml solrconfig.xml
```

Optionally add a JWT claim:
```
java -cp /opt/pontus/pontus-hbase-coprocessor-atlas-0.0.1-SNAPSHOT.jar JWTStoreClient solr '{ jwtClaim="/uk.manager/special" }'
```



Start, stop, and restart commands for Solr:
```
/opt/solr/bin/solr  start -cloud -p 8983 -h sandbox.hortonworks.com -z sandbox.hortonworks.com:2181/solr -s ~/solr-cores/core1 -a "-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=60000"
/opt/solr/bin/solr stop -cloud -p 8983 
/opt/solr/bin/solr  restart -cloud -p 8983 -h sandbox.hortonworks.com -z sandbox.hortonworks.com:2181/solr -s ~/solr-cores/core1 -a "-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=60000"
```



Creating new collection 'labs' using command:
```
curl -u : --negotiate 'http://localhost:8983/solr/admin/collections?action=CREATE&name=labs&numShards=1&replicationFactor=1&maxShardsPerNode=1&collection.configName=labs'
```

Or, if you need to delete it:
```
http://sandbox.hortonworks.com:8983/solr/admin/collections?action=DELETE&name=labs
```


## Troubleshooting:
  
  1) Symptom:
```
  939  ERROR (main) [   ] o.a.s.c.SolrCore null:org.apache.solr.common.SolrException: Error initializing kerberos authentication plugin: javax.servlet.ServletException: org.apache.hadoop.security.authentication.client.AuthenticationException: javax.security.auth.login.LoginException: No key to store
        at org.apache.solr.security.KerberosPlugin.init(KerberosPlugin.java:128)
        at org.apache.solr.core.CoreContainer.initializeAuthenticationPlugin(CoreContainer.java:292)
        at org.apache.solr.core.CoreContainer.load(CoreContainer.java:418)
```
        
Cure:      
        Check that kinit was not called by running klist; if anything similar to the following appears, delete the Ticket cache:
```        
        [solr@sandbox lib]$ klist
            Ticket cache: FILE:/tmp/krb5cc_1026
            Default principal: HTTP/sandbox.hortonworks.com@YOUR_REALM_GOES_HERE

            Valid starting     Expires            Service principal
            11/24/16 13:30:14  11/25/16 13:30:14  krbtgt/YOUR_REALM_GOES_HERE@YOUR_REALM_GOES_HERE
            renew until 11/24/16 13:30:14
      
        rm -f /tmp/krb5cc_`id -u`
```        
        Then restart the node:
```        
        /opt/solr/bin/solr  restart -cloud -p 8983 -h sandbox.hortonworks.com -z sandbox.hortonworks.com:2181/solr -s ~/solr-cores/core1 -a "-agentlib:jdwp=transport=dt_socket,server=y,suspend=n,address=60000"
```
        
  2) Symptom:      
```
      2217 ERROR (coreLoadExecutor-7-thread-1-processing-n:sandbox.hortonworks.com:8983_solr) [c:labs s:shard1 r:core_node1 x:core1] o.a.s.c.CoreContainer Error creating core [core1]: java.io.IOException: Login failure for hdfs-san
      dbox@YOUR_REALM_GOES_HERE from keytab /etc/security/keytabs/hdfs.headless.keytab: javax.security.auth.login.LoginException: Unable to obtain password from user

      java.lang.RuntimeException: java.io.IOException: Login failure for hdfs-sandbox@YOUR_REALM_GOES_HERE from keytab /etc/security/keytabs/hdfs.headless.keytab: javax.security.auth.login.LoginException: Unable to obtain password from user
```
Cure:
check permissions of keytab file.
        
        
   3) Symptom:
```   
         coreLoadExecutor-7-thread-1-processing-n:sandbox.hortonworks.com:8983_solr) [c:labs s:shard1 r:core_node1 x:core1] o.a.s.c.CoreContainer Error creating core [core1]: Index dir 'hdfs://sandbox.hortonworks.com/user/solr/labs/core_node1/core1/data/index/' of core 'core1' is already locked. The most likely cause is another Solr server (or another solr core in this server) also configured to use this directory; other possible causes may be specific to lockType: hdfs
      org.apache.solr.common.SolrException: Index dir 'hdfs://sandbox.hortonworks.com/user/solr/labs/core_node1/core1/data/index/' of core 'core1' is already locked. The most likely cause is another Solr server (or another solr core in this server) also configured to use this directory; other possible causes may be specific to lockType: hdfs
              at org.apache.solr.core.SolrCore.<init>(SolrCore.java:820)
              at org.apache.solr.core.SolrCore.<init>(SolrCore.java:658)
              at org.apache.solr.core.CoreContainer.create(CoreContainer.java:814)
              at org.apache.solr.core.CoreContainer.access$000(CoreContainer.java:87)
              at org.apache.solr.core.CoreContainer$1.call(CoreContainer.java:467)
              at org.apache.solr.core.CoreContainer$1.call(CoreContainer.java:458)
              at java.util.concurrent.FutureTask.run(FutureTask.java:266)
              at org.apache.solr.common.util.ExecutorUtil$MDCAwareThreadPoolExecutor$1.run(ExecutorUtil.java:231)
              at java.util.concurrent.ThreadPoolExecutor.runWorker(ThreadPoolExecutor.java:1142)
              at java.util.concurrent.ThreadPoolExecutor$Worker.run(ThreadPoolExecutor.java:617)
              at java.lang.Thread.run(Thread.java:745)
      Caused by: org.apache.lucene.store.LockObtainFailedException: Index dir 'hdfs://sandbox.hortonworks.com/user/solr/labs/core_node1/core1/data/index/' of core 'core1' is already locked. The most likely cause is another Solr server (or another solr core in this server) also configured to use this directory; other possible causes may be specific to lockType: hdfs
              at org.apache.solr.core.SolrCore.initIndex(SolrCore.java:525)
              at org.apache.solr.core.SolrCore.<init>(SolrCore.java:761)
              ... 
```              
              
Cure: 
remove the hdfs writelock  file  in hdfs://sandbox.hortonworks.com/user/solr/labs/core_node1/core1/data/index
       
or
       
ensure that he user has the correct permissions
       
       
  4) Symptom:
```
        Caused by: java.lang.SecurityException: java.io.IOException: Configuration Error:
        Line 13: expected [option key]
        at sun.security.provider.ConfigFile$Spi.<init>(ConfigFile.java:137)
        at sun.security.provider.ConfigFile.<init>(ConfigFile.java:102)
        at sun.reflect.NativeConstructorAccessorImpl.newInstance0(Native Method)
        at sun.reflect.NativeConstructorAccessorImpl.newInstance(NativeConstructorAccessorImpl.java:62)
        at sun.reflect.DelegatingConstructorAccessorImpl.newInstance(DelegatingConstructorAccessorImpl.java:45)
        at java.lang.reflect.Constructor.newInstance(Constructor.java:423)
        at java.lang.Class.newInstance(Class.java:442)
        at javax.security.auth.login.Configuration$2.run(Configuration.java:255)  
```       

Cure: 

ensure that you have a semi colon ONLY at the LAST entry within each {} section.

5) Symptom:
```
  2016-12-27 10:05:22.852 INFO  (qtp1295083508-25) [   ] o.a.s.s.SolrDispatchFilter Error authenticating
java.lang.NullPointerException
        at org.apache.hadoop.security.authentication.server.KerberosAuthenticationHandler$2.run(KerberosAuthenticationHandler.java:375)
        at org.apache.hadoop.security.authentication.server.KerberosAuthenticationHandler$2.run(KerberosAuthenticationHandler.java:347)
        at java.security.AccessController.doPrivileged(Native Method)
        at javax.security.auth.Subject.doAs(Subject.java:422)
        at org.apache.hadoop.security.authentication.server.KerberosAuthenticationHandler.authenticate(KerberosAuthenticationHandler.java:347)
        at org.apache.solr.security.KerberosPlugin$RequestContinuesRecorderAuthenticationHandler.authenticate(KerberosPlugin.java:551)
        at org.apache.hadoop.security.authentication.server.AuthenticationFilter.doFilter(AuthenticationFilter.java:519)
        at org.apache.solr.security.KerberosFilter.doFilter(KerberosFilter.java:60)
        at org.apache.solr.security.KerberosPlugin.doAuthenticate(KerberosPlugin.java:248)
        at org.apache.solr.servlet.SolrDispatchFilter.authenticateRequest(SolrDispatchFilter.java:362)
```
Cure:

Ensure that the browser says http://sandbox.hortonworks.com::8983/solr  
the exception above happens when the browser has http://localhost:8983/solr



    
